import json
from snappi_ixnetwork.timer import Timer


class Lag(object):
    """Transforms OpenAPI objects into IxNetwork objects
    - Lag to /lag
    Args
    ----
    - ixnetworkapi (Api): instance of the Api class
    """
    _ETHERNET = {
        'mac': {
            'ixn_attr': 'mac',
            'default': '_default_mac'
        },
        'mtu': {
            'ixn_attr': 'mtu',
            'default': '1500'
        },
    }
    
    _VLAN_TPID = {
        'x8100': 'ethertype8100',
        'x88a8': 'ethertype88a8',
        'x9100': 'ethertype9100',
        'x9200': 'ethertype9200',
        'x9300': 'ethertype9300'
    }
    
    _VLAN = {
        # 'tpid' : {
        #     'ixn_attr' : 'tpid',
        #     'default' : 'ethertype8100',
        #     'enum_map' : _VLAN_TPID
        # },
        'priority': {
            'ixn_attr': 'priority',
            'default': '0'
        },
        'vlanId': {
            'ixn_attr': 'vlanId',
            'default': '1'
        }
    }
    
    _LACP = {
        'actor_key': {
            'ixn_attr': 'actorKey',
            'default': '1'
        },
        'actor_port_number': {
            'ixn_attr': 'actorPortNumber',
            'default': '1'
        },
        'actor_port_priority': {
            'ixn_attr': 'actorPortPriority',
            'default': '1'
        },
        'actor_system_id': {
            'ixn_attr': 'actorSystemId',
            'default': '_default_system_id'
        },
        'actor_system_priority': {
            'ixn_attr': 'actorSystemPriority',
            'default': '1'
        },
        'lacpdu_periodic_time_interval': {
            'ixn_attr': 'lacpduPeriodicTimeInterval',
            'default': '0'
        },
        'lacpdu_timeout': {
            'ixn_attr': 'lacpduTimeout',
            'default': '0'
        },
        'actor_activity': {
            'ixn_attr': 'lacpActivity',
            'default': 'active'
        },
    }
    
    _STATIC = {
        'lag_id': {
            'ixn_attr': 'lagId',
            'default': '1'
        }
    }
    
    def __init__(self, ixnetworkapi):
        self._api = ixnetworkapi
        self._lag_ports = {}
    
    def config(self):
        """Transform config.ports into Ixnetwork.Vport
        1) delete any vport that is not part of the config
        2) create a vport for every config.ports[] that is not present in IxNetwork
        3) set config.ports[].location to /vport -location using resourcemanager
        4) set /vport/l1Config/... properties using the corrected /vport -type
        5) connectPorts to use new l1Config settings and clearownership
        """
        self._resource_manager = self._api._ixnetwork.ResourceManager
        self._ixn_lag = self._api._lag
        self._lag_ports = {}
        self._lags_config = self._api.snappi_config.lags
        with Timer(self._api, "Lag Configuration"):
            self._delete_lags()
            if len(self._lags_config) == 0:
                return
            self._create_lags()
        with Timer(self._api, "Lag Ethernet Configuration"):
            self._ethernet_config()
        with Timer(self._api, "Lag Protocol Configuration"):
            self._protocol_config()
    
    def _import(self, imports):
        if len(imports) > 0:
            errata = self._resource_manager.ImportConfig(
                json.dumps(imports), False)
            for item in errata:
                self._api.warning(item)
            return len(errata) == 0
        return True
    
    def _delete_lags(self):
        """Delete any Lags from the api server that do not exist in the new config
        """
        self._api._remove(self._ixn_lag, self._lags_config)
    
    def _select_lags(self):
        payload = {
            'selects': [{
                'from':
                    '/',
                'properties': [],
                'children': [{
                    'child':
                        'lag',
                    'properties': ['name'],
                    'filters': []
                }],
                'inlines': []
            }]
        }
        url = '%s/operations/select?xpath=true' % self._api._ixnetwork.href
        results = self._api._ixnetwork._connection._execute(url, payload)
        lags = {}
        if 'lag' in results[0]:
            for lag in results[0]['lag']:
                lags[lag['name']] = lag
        return lags
    
    def _select_vports(self):
        """Select all vports.
        Return them in a dict keyed by vport name.
        """
        payload = {
            'selects': [{
                'from':
                    '/',
                'properties': [],
                'children': [{
                    'child':
                        'vport',
                    'properties': [
                        'name',
                    ],
                    'filters': []
                }],
                'inlines': []
            }]
        }
        url = '%s/operations/select?xpath=true' % self._api._ixnetwork.href
        results = self._api._ixnetwork._connection._execute(url, payload)
        vports = {}
        if 'vport' in results[0]:
            for vport in results[0]['vport']:
                vports[vport['name']] = vport
        return vports
    
    def _get_vports(self, ixn_vports, ports):
        vports = []
        for port in ports:
            if port.port_name not in ixn_vports:
                raise Exception("Port {0} is not available".format(
                    port.name))
            vports.append(ixn_vports[port.port_name]['xpath'])
        return vports
    
    def _create_lags(self):
        """Add any Lags to the api server that do not already exist
        """
        self._ixn_lag.find()
        existing_lags = [ixn_lag.Name
                         for ixn_lag in self._ixn_lag]
        for lag in self._lags_config:
            self._lag_ports[lag.name] = lag.ports
            if lag.name not in existing_lags:
                self._ixn_lag.add(Name=lag.name)
        imports = []
        ixn_vports = self._select_vports()
        ixn_lags = self._select_lags()
        for name, ixn_lag in ixn_lags.items():
            self._api.ixn_objects[name] = ixn_lag['href']
            lag_import = {
                'xpath': ixn_lag['xpath'],
                'vports': self._get_vports(ixn_vports,
                                           self._lag_ports[name])
            }
            imports.append(lag_import)
        self._import(imports)
    
    def _set_multivalue(self, source, attribute, value):
        xpath = "/multivalue[@source = '{0} {1}']".format(
            source, attribute)
        if isinstance(value, list):
            return {
                "xpath": "{0}/valueList".format(xpath),
                "values": value
            }
        else:
            return {
                "xpath": "{0}/singleValue".format(xpath),
                "value": value
            }
    
    def _ethernet_config(self):
        imports = []
        ixn_lags = self._select_lags()
        for name, ports in self._lag_ports.items():
            ether_xpath = "{0}/protocolStack/ethernet[1]".format(
                ixn_lags[name]['xpath'])
            ixn_lag = self._ixn_lag.find(Name='^%s$' % self._api.special_char(
                                            name))
            ixn_proto_stack = ixn_lag.ProtocolStack.find()
            if len(ixn_proto_stack) == 0:
                ixn_proto_stack = ixn_lag.ProtocolStack.add()[0]
            ixn_proto_stack.Multiplier = 1
            for eth_attr in Lag._ETHERNET:
                attr_values = self._configure_attribute(eth_attr, Lag._ETHERNET,
                                                        ports, 'ethernet')
                imports.append(self._set_multivalue(ether_xpath, attr_values.ixn_attribute,
                                                    attr_values.config_value))
            vlans = self._process_vlans(ports)
            if len(vlans) > 0:
                if len(vlans) > 0:
                    imports.append(self._set_multivalue(ether_xpath, 'enableVlans',
                                                        True))
                    imports.append({
                        "xpath": ether_xpath,
                        "vlanCount": len(vlans)
                    })
                
                for idx, vlan in enumerate(vlans):
                    for vlan_attr in Lag._VLAN:
                        attr_values = self._configure_attribute(vlan_attr, Lag._VLAN,
                                                                vlan)
                        vlan_xpath = "{0}/vlan[{1}]".format(ether_xpath, idx + 1)
                        imports.append(self._set_multivalue(vlan_xpath, attr_values.ixn_attribute,
                                                            attr_values.config_value))
        self._import(imports)
    
    def _protocol_config(self):
        imports = []
        ixn_lags = self._select_lags()
        for name, ports in self._lag_ports.items():
            ixn_lag = self._ixn_lag.find(Name='^%s$' % self._api.special_char(
                name))
            ixn_eth = ixn_lag.ProtocolStack.find().Ethernet.find()
            ixn_lacp = ixn_eth.Lagportlacp.find()
            ixn_static = ixn_eth.Lagportstaticlag.find()
            choice = None
            protocols = []
            for port in ports:
                protocol = port.protocol
                if choice is None:
                    choice = protocol.choice
                elif choice != protocol.choice:
                    raise Exception("Please configure same protocol "
                                    "[static, lacp] within same Lag ports")
                if choice == 'lacp':
                    protocol.lacp.actor_system_id = protocol.lacp.actor_system_id.replace(':', ' ')
                protocols.append(protocol)
            if choice is None:
                if len(ixn_static) > 0:
                    ixn_static.Active.Single(False)
                if len(ixn_lacp) > 0:
                    ixn_lacp.Active.Single(False)
                return
            if choice == 'lacp':
                if len(ixn_static) > 0:
                    ixn_static.remove()
                lacp_xpath = "{0}/protocolStack/ethernet[1]/lagportlacp[1]".format(
                    ixn_lags[name]['xpath'])
                imports.append(self._set_multivalue(lacp_xpath, 'active', True))
                for lacp_attr in Lag._LACP:
                    attr_values = self._configure_attribute(lacp_attr, Lag._LACP,
                                                            protocols, choice)
                    imports.append(self._set_multivalue(lacp_xpath, attr_values.ixn_attribute,
                                                        attr_values.config_value))
            else:
                if len(ixn_lacp) > 0:
                    ixn_lacp.remove()
                static_xpath = "{0}/protocolStack/ethernet[1]/lagportstaticlag[1]".format(
                    ixn_lags[name]['xpath'])
                imports.append(self._set_multivalue(static_xpath, 'active', True))
                for static_attr in Lag._STATIC:
                    attr_values = self._configure_attribute(static_attr, Lag._STATIC,
                                                            protocols, choice)
                    imports.append(self._set_multivalue(static_xpath, attr_values.ixn_attribute,
                                                        attr_values.config_value))
        
        self._import(imports)
    
    def _process_vlans(self, ports):
        vlan_count = -1
        vlan_list = []
        for port in ports:
            vlans = port.ethernet.vlans
            if vlan_count == -1:
                vlan_count = len(vlans)
                for i in range(len(vlans)):
                    vlan_list.append([vlans[i]])
            elif vlan_count == len(vlans):
                for i in range(len(vlans)):
                    vlan_list[i].append(vlans[i])
            else:
                raise Exception("Please configure equal numbers of VLANs")
        return vlan_list
    
    def _configure_attribute(self, attr, mapper, parent_list, obj_name=None):
        attr_values = ProtocolAttributes()
        attr_mapper = mapper[attr]
        ixn_attr = attr_mapper['ixn_attr']
        default_value = attr_mapper['default']
        enum_map = attr_mapper.get('enum_map')
        default_obj = getattr(self, default_value, None)
        attr_values.ixn_attribute = ixn_attr
        for parent in parent_list:
            if obj_name is not None:
                parent = getattr(parent, obj_name, None)
            config_value = getattr(parent, attr, None)
            if config_value is not None:
                if enum_map is None:
                    attr_values.config_value = str(config_value)
                else:
                    attr_values.config_value = enum_map[
                        str(config_value)]
            elif default_obj is None:
                attr_values.config_value = default_value
            else:
                attr_values.config_value = default_obj()
        return attr_values
    
    def _select_protcols(self, lag_href):
        payload = {
            'selects': [{
                'from':
                    lag_href,
                'properties': [],
                'children': [{
                    'child':
                        'protocolStack',
                    'properties': ['*'],
                    'filters': []
                }, {
                    'child':
                        'lagMode',
                    'properties': ['*'],
                    'filters': []
                }],
                'inlines': []
            }]
        }
        url = '%s/operations/select?xpath=true' % self._api._ixnetwork.href
        results = self._api._ixnetwork._connection._execute(url, payload)
        lags = {}
        if 'lag' in results[0]:
            for lag in results[0]['lag']:
                lags[lag['name']] = lag
        return lags
    
    def _set_result_value(self, row, column_name, column_value, column_type=str):
        if len(self._column_names) > 0 and column_name not in self._column_names:
            return
        try:
            row[column_name] = column_type(column_value)
        except:
            if column_type.__name__ in ['float', 'int']:
                row[column_name] = 0
            else:
                row[column_type] = column_value
    
    def results(self, request):
        """Return port results
        """
        if request.column_names is None:
            self._column_names = []
        else:
            self._column_names = request.column_names
        port_rows = {}
        for vport in self._api.select_vports().values():
            port_row = {}
            self._set_result_value(port_row, 'name', vport['name'])
            location = vport['location']
            if vport['connectionState'].startswith('connectedLink') is True:
                location += ';connected'
            elif len(location) > 0:
                location += ';' + vport['connectionState']
            else:
                location = vport['connectionState']
            self._set_result_value(port_row, 'location', location)
            self._set_result_value(port_row, 'link', 'up' if vport['connectionState'] == 'connectedLinkUp' else 'down')
            self._set_result_value(port_row, 'capture', 'stopped')
            port_rows[vport['name']] = port_row
        try:
            table = self._api.assistant.StatViewAssistant('Port Statistics')
            for row in table.Rows:
                port_row = port_rows[row['Port Name']]
                self._set_result_value(port_row, 'frames_tx', row['Frames Tx.'], int)
                self._set_result_value(port_row, 'frames_rx', row['Valid Frames Rx.'], int)
                self._set_result_value(port_row, 'frames_tx_rate', row['Frames Tx. Rate'], float)
                self._set_result_value(port_row, 'frames_rx_rate', row['Valid Frames Rx. Rate'], float)
                self._set_result_value(port_row, 'bytes_tx', row['Bytes Tx.'], int)
                self._set_result_value(port_row, 'bytes_rx', row['Bytes Rx.'], int)
                self._set_result_value(port_row, 'bytes_tx_rate', row['Bytes Tx. Rate'], float)
                self._set_result_value(port_row, 'bytes_rx_rate', row['Bytes Rx. Rate'], float)
                self._set_result_value(port_row, 'pfc_class_0_frames_rx', row['Rx Pause Priority Group 0 Frames'], int)
                self._set_result_value(port_row, 'pfc_class_1_frames_rx', row['Rx Pause Priority Group 1 Frames'], int)
                self._set_result_value(port_row, 'pfc_class_2_frames_rx', row['Rx Pause Priority Group 2 Frames'], int)
                self._set_result_value(port_row, 'pfc_class_3_frames_rx', row['Rx Pause Priority Group 3 Frames'], int)
                self._set_result_value(port_row, 'pfc_class_4_frames_rx', row['Rx Pause Priority Group 4 Frames'], int)
                self._set_result_value(port_row, 'pfc_class_5_frames_rx', row['Rx Pause Priority Group 5 Frames'], int)
                self._set_result_value(port_row, 'pfc_class_6_frames_rx', row['Rx Pause Priority Group 6 Frames'], int)
                self._set_result_value(port_row, 'pfc_class_7_frames_rx', row['Rx Pause Priority Group 7 Frames'], int)
        except:
            pass
        return port_rows.values()


class ProtocolAttributes(object):
    def __init__(self):
        self._ixn_attr = None
        self._config_value = []
    
    @property
    def ixn_attribute(self):
        return self._ixn_attr
    
    @ixn_attribute.setter
    def ixn_attribute(self, value):
        self._ixn_attr = value
    
    @property
    def config_value(self):
        return self._config_value
    
    @config_value.setter
    def config_value(self, value):
        self._config_value.append(value)


from snappi_ixnetwork.timer import Timer
import time


class ProtocolMetrics(object):
    # Currently paging of the statistic view is not handled
    # TODO Need to enhance when device groups statistics reach
    # more than one page.

    _SUPPORTED_PROTOCOLS_ = [
        "bgpv4", "bgpv6"
    ]

    _TOPO_STATS = {
        "name": "name",
        "total": "sessions_total",
        "up": "sessions_up",
        "down": "sessions_down",
        "notStarted": "sessions_not_started"
    }

    _RESULT_COLUMNS = {
        "bgpv4": [
            ("name", "Device Group", str),
            ("sessions_total", "Sessions Total", int),
            ("sessions_up", "Sessions Up", int),
            ("sessions_down", "Sessions Down", int),
            ("sessions_not_started", "Sessions Not Started", int),
            ("routes_advertised", "Routes Advertised", int),
            ("routes_withdrawn", "Routes Withdrawn", int),
        ],
        "bgpv6": [
            ("name", "Device Group", str),
            ("sessions_total", "Sessions Total", int),
            ("sessions_up", "Sessions Up", int),
            ("sessions_down", "Sessions Down", int),
            ("sessions_not_started", "Sessions Not Started", int),
            ("routes_advertised", "Routes Advertised", int),
            ("routes_withdrawn", "Routes Withdrawn", int),
        ]
    }

    _PROTO_NAME_MAP_ = {
        "bgpv4": {
            'per_port': 'BGP Peer Per Port',
            'drill_down': 'BGP Peer Drill Down',
            'drill_down_options': [
                'BGP Peer:Per Device Group', 'BGP Peer:Per Session'
            ],
            'supported_stats': [s[0] for s in _RESULT_COLUMNS["bgpv4"]],
            'ixn_name': 'bgpIpv4Peer'
        },
        "bgpv6": {
            'per_port': 'BGP\+ Peer Per Port',
            'drill_down': 'BGP\+ Peer Drill Down',
            'drill_down_options': [
                'BGP+ Peer:Per Device Group', 'BGP+ Peer:Per Session'
            ],
            'supported_stats': [s[0] for s in _RESULT_COLUMNS["bgpv6"]],
            'ixn_name': 'bgpIpv6Peer'
        }
    }

    def __init__(self, ixnetworkapi):
        self._api = ixnetworkapi
        self.ixn = None
        self.columns = []
        self.device_names = []
        self.metric_timeout = 90
        self.interval = 1

    def _get_search_payload(self, parent, child, properties, filters):
        payload = {
            'selects': [{
                'from': parent,
                'properties': [],
                'children': [{
                    'child': child,
                    'properties': properties,
                    'filters': filters
                }],
                'inlines': []
            }]
        }
        url = '{}/operations/select?xpath=true'.format(
            self.ixn.href
        )
        return (url, payload)

    def _select_view(self, view_filters=[]):
        url, payload = self._get_search_payload(
            '/statistics', 'view', ['caption'], view_filters
        )
        result = self.ixn._connection._execute(url, payload)[0]
        return result.get('view')

    def _wait_for(self, func, exp_msg, interval, timeout):
        end_time = round(time.time()) + timeout
        while True:
            res = func()
            if round(time.time()) >= end_time:
                raise Exception(exp_msg)
            if res is not None:
                return res
            time.sleep(interval)

    def get_supported_protocols(self):
        """
        Return the protocols that are supported currently
        """
        return self._SUPPORTED_PROTOCOLS_

    def _port_list_in_per_port(self, protocol):
        self.ixn = self._api.assistant._ixnetwork
        protocol_name = self._PROTO_NAME_MAP_.get(protocol)
        if protocol_name is None:
            raise NotImplementedError(
                "{} is Not Implemented".format(protocol)
            )
        filter = [{
            'property': 'caption',
            'regex': '^%s$' % protocol_name['per_port']
        }]
        view = self._wait_for(
            lambda: self._select_view(filter),
            "could not retrieve the view for %s" % protocol,
            self.interval,
            self.metric_timeout
        )
        res = self._get_column_values(view[0]['href'], 'Port')
        if res is None or len(res) == 0:
            raise Exception("Could not retrieve the Port column")
        return (res, view[0])

    def _get_column_values(self, href, value):
        payload = {
            'arg1': href,
            'arg2': value
        }
        url = '{}/statistics/view/operations/getcolumnvalues'.format(
            self.ixn.href
        )
        res = self._api._request('POST', url, payload)
        return res.get('result')

    def _get_value(self, href, row_label, column_label):
        payload = {
            'arg1': href,
            'arg2': row_label,
            'arg3': column_label
        }
        url = '{}/statistics/view/operations/getvalue'.format(
            self.ixn.href
        )
        res = self._api._request('POST', url, payload)
        return res.get('result')

    def _check_if_page_ready(self, view):
        count = 0
        while True:
            view.Refresh()
            if view.Data.IsReady:
                break
            if count >= self.metric_timeout:
                raise Exception("View Page is not ready")
            time.sleep(0.5)
            count += 1

    def _port_names_from_devices(self):
        config = self._api.snappi_config
        if len(self.device_names) == 0:
            port_list = [p.name for p in config.ports]
            return port_list
        port_list = [
            d.container_name
            for d in config.devices if d.name in self.device_names
        ]
        return port_list

    def _do_drill_down(self, view, per_port, row_index, drill_option):
        url, payload = self._get_search_payload(
            view['href'], 'drillDown', [
                'targetDrillDownOption', 'targetRowIndex'
            ], []
        )
        result = self.ixn._connection._execute(url, payload)[0]
        if result.get('drillDown') is None:
            raise Exception('Could not fetch drill down node')

        payload = {
            'targetDrillDownOption': drill_option,
            'targetRowIndex': row_index,
        }
        url = result['drillDown']['href']
        count = 0
        while count < 5:
            # retrying as the linux api server throws error for first 2
            # consecutive executions
            try:
                self._api._request('PATCH', url, payload)
                break
            except Exception:
                self._check_if_page_ready(
                    self.ixn.Statistics.View.find(Caption=per_port)
                )
                time.sleep(0.3)
                count += 1

        url = '{}/statistics/view/drillDown/operations/dodrilldown'.format(
            self.ixn.href
        )
        payload = {
            'arg1': result['drillDown']['href']
        }
        self._api._request('POST', url, payload)
        return

    def _get_per_device_group_stats(self, protocol):
        ports, v = self._port_list_in_per_port(protocol)
        config_ports = self._port_names_from_devices()
        indices = [
            ports.index(p) for p in list(set(config_ports)) if p in ports
        ]
        drill_option = self._PROTO_NAME_MAP_[protocol]['drill_down_options'][0]
        drill_name = self._PROTO_NAME_MAP_[protocol]['drill_down']
        per_port = self._PROTO_NAME_MAP_[protocol]['per_port']
        column_names = self._RESULT_COLUMNS.get(protocol)
        row_lst = list()
        for i in indices:
            try:
                drill = self.ixn.Statistics.View.find(Caption=drill_name)
                self._do_drill_down(v, per_port, i, drill_option)
                self._check_if_page_ready(drill)
            except Exception as e:
                msg = """
                Could not retrive drill down view \
                at row index {} {}""".format(i, e)
                raise Exception(msg)
            dev_names = self._get_column_values(
                drill.href, 'Device Group'
            )
            for index, dev_name in enumerate(dev_names):
                if len(self.device_names) > 0 and\
                        dev_name not in self.device_names:
                    continue
                row_dt = dict()
                for sn, ixn, typ in column_names:
                    try:
                        value = self._get_value(
                            drill.href, dev_name, ixn
                        )
                    except Exception:
                        value = 'NA'
                    self._set_result_value(
                        row_dt, sn, value, typ
                    )
                row_lst.append(row_dt)
        return row_lst

    def _set_result_value(self, row_dt, stat_name, stat_value, stat_type=str):
        if len(self.columns) == 0 or stat_name in self.columns:
            try:
                row_dt[stat_name] = stat_type(stat_value)
            except Exception:
                row_dt[stat_name] = 0 \
                    if stat_type.__name__ in ['float', 'int'] else stat_value

    def _topo_stats(self, protocol):
        url = '%s/operations/gettopologystatus' % self.ixn.href
        res = self._api._request('POST', url)
        stats_map = {}
        for d in res['result']:
            if self._PROTO_NAME_MAP_[protocol]['ixn_name'] not in d['arg1']:
                continue
            stats_map[d['arg1']] = {
                self._TOPO_STATS[i['arg1']]: i['arg2'] for i in d['arg2']
            }
        myfilter = [{
            'property': 'name',
            'regex': '.*' if len(self.device_names) == 0
                        else '^%s$' % "|".join(
                            self._api.special_char(self.device_names)
                            )
        }]
        url, payload = self._get_search_payload(
            '/topology',
            '(?i)^(deviceGroup)$',
            ['name'],
            myfilter
        )
        result = self.ixn._connection._execute(url, payload)
        for t in result:
            if t.get('deviceGroup') is None:
                continue
            for d in t['deviceGroup']:
                for p in stats_map:
                    if d['href'] not in p:
                        continue
                    stats_map[p]['name'] = d['name']
        if self.device_names == []:
            return [stats_map[p] for p in stats_map]
        return [
            stats_map[p]
            for p in stats_map
            if stats_map['name'] in self.device_names
        ]

    def _filter_stats(self, protocol):
        self.ixn = self._api.assistant._ixnetwork
        f = [
            i for i in filter(
                lambda y: y not in self._TOPO_STATS.values(),
                self.columns
            )]
        if len(f) > 0:
            return self._get_per_device_group_stats(protocol)
        else:
            return self._topo_stats(protocol)

    def results(self, request):
        """
        Return the Protocol statistics
        args:
            request: <type obj> MetricRequest object
            protocol: <type str> protocol name
        *Note: supported protocols can be fetched via,
        self.get_supported_protocols()
        """
        protocol = request.choice
        request = getattr(request, protocol)
        self.device_names = request._properties.get('device_names')
        if self.device_names is None:
            self.device_names = []
        self.columns = request._properties.get('column_names')
        if self.columns is None or self.columns == []:
            self.columns = self._PROTO_NAME_MAP_[protocol][
                'supported_stats'
            ]
        if len(self.columns) > 0:
            self.columns.append('name')
            self.columns = list(set(self.columns))
        with Timer(self._api, "Fetching {} Metrics".format(protocol)):
            return self._filter_stats(protocol)




class Events(object):
    def __init__(self, ixnetworkapi):
        self._api = ixnetworkapi
    
    def config(self):
        events = self._api.snappi_config._properties.get('events')
        if events is None:
            return
        ixn_CpdpConvergence = self._api._traffic.Statistics.CpdpConvergence
        if self.event_allow('rx_rate_threshold'):
            if self._api.traffic_item.has_latency is True:
                raise Exception("We are supporting either latency or rx_rate_threshold")
            ixn_CpdpConvergence.Enabled = True
            ixn_CpdpConvergence.EnableControlPlaneEvents = True
            ixn_CpdpConvergence.EnableDataPlaneEventsRateMonitor = True
            if events.rx_rate_threshold.threshold is not None:
                ixn_CpdpConvergence.DataPlaneThreshold = events.rx_rate_threshold.threshold
        else:
            ixn_CpdpConvergence.Enabled = False
    
    def event_allow(self, event_name):
        events = self._api.snappi_config._properties.get('events')
        if events is None:
            return False
        event = events._properties.get(event_name)
        if events.enable is True or (event is not None and
                             event.enable is True):
            return True
        else:
            return False


import re
import time
from collections import namedtuple
from snappi_ixnetwork.exceptions import SnappiIxnException
from snappi_ixnetwork.timer import Timer
try:
    import snappi_convergence
except Exception:
    raise SnappiIxnException(500, "snappi_convergence not installed")
from snappi_ixnetwork.ixnetworkapi import Api as snappiApi

# todo: api should remove when package will ready
def api(location = None, ext='ixnetwork'):
    return Api(location= location)


class Api(snappi_convergence.Api):
    _CONVERGENCE = {
        ('data_plane_convergence_us', 'DP/DP Convergence Time (us)', float),
        ('control_plane_data_plane_convergence_us', 'CP/DP Convergence Time (us)', float),
    }
    
    _EVENT = {
        ('begin_timestamp_ns', 'Event Start Timestamp', int),
        ('end_timestamp_ns', 'Event End Timestamp', int)
    }
    
    def __init__(self,
                 **kwargs):
        
        self._convergence_timeout = 10
        self._api = snappiApi(**kwargs)
        self._event_info = None
        
    def set_config(self, payload):
        try:
            cvg_config = self.convergence_config()
            if isinstance(payload, (type(cvg_config),
                                    str)) is False:
                raise TypeError(
                    'The content must be of type Union[Config, str]')

            if isinstance(payload, str) is True:
                payload = cvg_config.deserialize(payload)
            config = payload._properties.get('config')
            if config is None:
                raise Exception("config should not None")
            self._api.config_ixnetwork(config)
            rx_rate_threshold = payload._properties.get('rx_rate_threshold')
            ixn_CpdpConvergence = self._api._traffic.Statistics.CpdpConvergence
            if rx_rate_threshold is not None:
                if self._api.traffic_item.has_latency is True:
                        raise Exception("We are supporting either latency or rx_rate_threshold")
                ixn_CpdpConvergence.Enabled = True
                ixn_CpdpConvergence.EnableControlPlaneEvents = True
                ixn_CpdpConvergence.EnableDataPlaneEventsRateMonitor = True
                ixn_CpdpConvergence.DataPlaneThreshold = rx_rate_threshold
            else:
                ixn_CpdpConvergence.Enabled = False
            for ixn_traffic_item in self._api._traffic_item.find():
                ixn_traffic_item.Tracking.find()[0].TrackBy = ['destEndpoint0',
                                                               'destSessionDescription0']
        except Exception as err:
            raise SnappiIxnException(err)
        app_errors = self._api._globals.AppErrors.find()
        bad_requests = []
        if len(app_errors) > 0:
            current_errors = app_errors[0].Error.find()
            if len(current_errors) > 0:
                for error in current_errors:
                    if error.Name == 'JSON Import Issues':
                        bad_requests = [instance.SourceValues[0]
                                        for instance in error.Instance.find()]
        if bad_requests:
            if len(bad_requests) == 1:
                raise SnappiIxnException(400, "Bad request error: {}".format(
                    bad_requests[0]))
            raise SnappiIxnException(400, "Bad request errors:", bad_requests)
        return self._request_detail()
    
    def set_state(self, payload):
        try:
            cvg_state = self.convergence_state()
            if isinstance(payload, (type(cvg_state),
                                    str)) is False:
                raise TypeError(
                    'The content must be of type Union[LinkState, str]')
            if isinstance(payload, str):
                payload = cvg_state.deserialize(payload)
            self._api._connect()
            if payload.choice is None:
                raise Exception("state [transmit/ link/ route] must configure")
            event_names = []
            event_state = None
            event_type = payload.choice
            EventInfo = namedtuple("EventInfo", ["event_type",
                                                 "event_state",
                                                 "event_names"])
            if payload.choice == 'transmit':
                transmit = payload.transmit
                self._api.traffic_item.transmit(transmit)
            elif payload.choice == 'link':
                link = payload.link
                if link.port_names is not None:
                    event_names = link.port_names
                    event_state = link.state
                    self._api.vport.set_link_state(link)
            elif payload.choice == 'route':
                route = payload.route
                event_state = route.state
                with Timer(self._api, 'Setting route state'):
                    event_names = self._api.ngpf.set_route_state(route)
            else:
                raise Exception("These[transmit/ link/ route] are valid convergence_state")
            self._event_info = EventInfo(event_type,
                                         event_state,
                                         event_names)
        except Exception as err:
            raise SnappiIxnException(err)
        return self._request_detail()
    
    def get_results(self, payload):
        try:
            self._api._connect()
            cvg_req = self.convergence_request()
            if isinstance(payload, (type(cvg_req),
                                    str)) is False:
                raise TypeError(
                    'The content must be of type Union[MetricsRequest, str]')
            if isinstance(payload, str) is True:
                payload = cvg_req.deserialize(payload)
            if payload.choice is None:
                raise Exception("state [metrics/ convergence] must configure")
            cvg_res = self.convergence_response()
            if payload.choice == 'metrics':
                response = self._api.traffic_item.results(
                            payload.metrics)
                cvg_res.flow_metric.deserialize(response)
            elif payload.choice == 'convergence':
                response = self._result(
                            payload.convergence)
                cvg_res.flow_convergence.deserialize(response)
            else:
                raise Exception("These[metrics/ convergence] are valid convergence_request")
            return cvg_res
        except Exception as err:
            raise SnappiIxnException(err)

    def _request_detail(self):
        request_detail = self.response_warning()
        errors = self._api._errors
        warnings = list()
        app_errors = self._api._globals.AppErrors.find()
        if len(app_errors) > 0:
            current_errors = app_errors[0].Error.find()
            if len(current_errors) > 0:
                for error in current_errors:
                    # Add loop as sequence are not sorted
                    match = [o for o in self._api._ixn_errors if o.Name == error.Name
                                                    and o.LastModified == error.LastModified]
                    if len(match) == 0:
                        if error.ErrorLevel == 'kWarning':
                            warnings.append("IxNet - {0}".format(error.Description))
                        if error.ErrorLevel == 'kError':
                            errors.append("IxNet - {0}".format(error.Description))
        # request_detail.errors = errors
        if len(errors) > 0:
            self._api._errors = []
            raise SnappiIxnException(500, errors)
        request_detail.warnings = warnings
        return request_detail
    
    def _result(self, request):
        flow_names = request._properties.get('flow_names')
        if not isinstance(flow_names, list):
            msg = "Invalid format of flow_names passed {},\
                                                expected list".format(flow_names)
            raise Exception(msg)
        if flow_names is None or len(flow_names) == 0:
            flow_names = [flow.name for flow in self._api.snappi_config.flows]
        flow_stat = self._api.assistant.StatViewAssistant(
            'Flow Statistics')
        flow_rows = flow_stat.Rows
        traffic_stat = self._api.assistant.StatViewAssistant(
            'Traffic Item Statistics')
        traffic_index = {}
        drill_down_option = 'Drill down per Dest Endpoint'
        for index, row in enumerate(self._get_traffic_rows(
                    traffic_stat, drill_down_option)):
            traffic_index[row['Traffic Item']] = index
        response = []
        for flow_name in flow_names:
            convergence = {
                'name': flow_name
            }
            if flow_name not in traffic_index.keys():
                raise Exception("Somehow flow %s is missing" % flow_name)
            
            drill_down_options = traffic_stat.DrillDownOptions()
            drilldown_index = drill_down_options.index(drill_down_option)
            drill_down = traffic_stat.Drilldown(traffic_index[flow_name], drill_down_option,
                                                traffic_stat.TargetRowFilters()[drilldown_index])
            drill_down_result = drill_down.Rows[0]
            for external_name, internal_name, external_type in self._CONVERGENCE:
                self._set_result_value(convergence, external_name, drill_down_result[
                    internal_name], external_type)
            
            events = []
            interruption_time = None
            for flow_result in flow_rows:
                if flow_result['Traffic Item'] != flow_name:
                    continue
                convergence['port_tx'] = flow_result['Tx Port']
                convergence['port_rx'] = flow_result['Rx Port']
                event_name = flow_result['Event Name']
                if event_name == '':
                    continue
                event = self._get_event(event_name, flow_result)
                for external_name, internal_name, external_type in self._EVENT:
                    value = int(flow_result[internal_name].split('.')[-1])
                    self._set_result_value(event, external_name, value * 1000, external_type)
                events.append(event)
                if interruption_time is None:
                    interruption_time = float(flow_result['DP Above Threshold Timestamp'].split(':')[-1]) - \
                                        float(flow_result['DP Below Threshold Timestamp'].split(':')[-1])
                    self._set_result_value(convergence, 'service_interruption_time_ns',
                                           interruption_time, float)
            convergence['events'] = events
            response.append(convergence)
        return response
    
    def _get_event(self, event_name, flow_result):
        event = {}
        if re.search(r'Port Link Up', event_name):
            if flow_result['Tx Port'] in self._event_info.event_names:
                event['source'] = flow_result['Tx Port']
            elif flow_result['Rx Port'] in self._event_info.event_names:
                event['source'] = flow_result['Rx Port']
            else:
                self._api.warning("Not find any event source")
                event['source'] = ''
            if self._event_info.event_state == 'up':
                event['type'] = 'link_up'
            else:
                event['type'] = 'link_down'
        else:
            for route_name in self._api.ixn_route_objects.keys():
                if re.search(route_name,
                             event_name) is not None:
                    event['source'] = route_name
                    event_type = event_name.split(route_name)[-1]
                    if event_type.strip().lower() == 'disable':
                        event_type = 'route_withdraw'
                    else:
                        event_type = "route_advertise"
                    event['type'] = event_type
                    break
        return event
    
    def _set_result_value(self,
                          row,
                          column_name,
                          column_value,
                          column_type=str):
        try:
            row[column_name] = column_type(column_value)
        except:
            if column_type.__name__ in ['float', 'int']:
                row[column_name] = 0
            else:
                row[column_type] = column_value

    def _get_traffic_rows(self, traffic_stat, drill_down_option):
        count = 0
        sleep_time = 0.5
        while True:
            has_event = False
            drill_down_options = traffic_stat.DrillDownOptions()
            if drill_down_option in drill_down_options:
                has_event = True
                break
            if has_event is True:
                break
            if count * sleep_time > self._convergence_timeout:
                raise Exception("Somehow \'Drill down per Dest Endpoint\' not available")
            time.sleep(sleep_time)
            count += 1
        return traffic_stat.Rows


"""Build distributions

To build `python setup.py sdist --formats=gztar bdist_wheel --universal`
"""
import os
from setuptools import setup
import requests
import json


RELEASES_URL = (
    'https://api.github.com/repos/open-traffic-generator/models/releases'
)

response = requests.request('GET', RELEASES_URL, allow_redirects=True)
assert response.status_code == 200
releases = json.loads(response.content)
# get latest release from v0.1.x branch
MODELS_RELEASES = [r['tag_name'] for r in releases if 'v0.1.' in r['tag_name']]

OPENAPI_URL = (
    'https://github.com/open-traffic-generator/models/releases/download/%s'
    '/openapi.yaml'
) % MODELS_RELEASES[0]

response = requests.request('GET', OPENAPI_URL, allow_redirects=True)
assert response.status_code == 200

# put the downloaded file inside docs dir of package
doc_dir = './snappi_ixnetwork/docs'
if os.path.exists(doc_dir) is False:
    os.mkdir(doc_dir)
with open(os.path.join(doc_dir, 'openapi.yaml'), 'wb') as fp:
    fp.write(response.content)

# read long description and version number
pkg_name = 'snappi_ixnetwork'
base_dir = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(base_dir, 'README.md')) as fid:
    long_description = fid.read()
with open(os.path.join(base_dir, 'VERSION')) as fid:
    version_number = fid.read()

setup(
    name=pkg_name,
    version=version_number,
    description='The IxNetwork Open Traffic Generator Python Package',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/open-traffic-generator',
    author='ajbalogh',
    author_email='andy.balogh@keysight.com',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Testing :: Traffic Generation',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3'
    ],
    keywords='ixnetwork testing open traffic generator automation',
    packages=[pkg_name],
    include_package_data=True,
    python_requires='>=2.7, <4',
    install_requires=[
        'ixnetwork-restpy>=1.0.52'
    ],
    extras_require={
        'dev': [
            'snappi==0.4.0',
            'snappi_convergence==0.0.5',
            'pytest',
            'flake8==3.8.4',
            'dpkt==1.9.4',
        ]
    }
)



import json
import time
import io
from snappi_ixnetwork.timer import Timer

class Capture(object):
    """Transforms OpenAPI objects into IxNetwork objects

    Args
    ----
    - ixnetworkapi (Api): instance of the Api class

    Transformations
    ---------------
    - /components/schemas/Capture to /vport/capture

    Process
    -------
    - Configure capture according to Filter

    Notes
    -----

    """
    _IPV4_OFFSET_MAP = {
        'version': '14',
        'headeer_length': '14',
        'priority': '15',
        'total_length': '16',
        'identification': '18',
        'reserved': '20',
        'dont_fragment': '20',
        'more_fragments': '20',
        'fragment_offset': '20',
        'time_to_live': '22',
        'protocol': '23',
        'header_checksum': '24',
        'src': '26',
        'dst': '30',
    }
    
    _VLAN_OFFSET_MAP = {
        'priority': '34',
        'cfi': '34',
        'id': '34',
        'protocol': '36',
    }
    
    def __init__(self, ixnetworkapi):
        self._api = ixnetworkapi
        self._capture_request = None
    
    def _import(self, imports):
        if len(imports) > 0:
            errata = self._resource_manager.ImportConfig(
                json.dumps(imports), False)
            for item in errata:
                self._api.warning(item)
            return len(errata) == 0
        return True
    
    def config(self):
        """Overwrite any capture settings
        """
        self._resource_manager = self._api._ixnetwork.ResourceManager
        imports = []
        vports = self._api.select_vports()
        for vport in vports.values():
            if vport['capture']['hardwareEnabled'] is True or vport['capture'][
                    'softwareEnabled'] is True:
                capture = {
                    'xpath': vport['capture']['xpath'],
                    'captureMode': 'captureTriggerMode',
                    'hardwareEnabled': False,
                    'softwareEnabled': False
                }
                imports.append(capture)
        for capture_item in self._api.snappi_config.captures:
            if capture_item.port_names is None:
                continue
            for port_name in capture_item.port_names:
                capture_mode = 'captureTriggerMode'
                if capture_item.overwrite:
                    capture_mode = 'captureContinuousMode'
                    reset = {'xpath': vports[port_name]['xpath'] + '/capture/trigger'}
                    reset['captureTriggerEnable'] = False
                    self._import(reset)
                capture = {
                    'xpath': vports[port_name]['xpath'] + '/capture',
                    'captureMode': capture_mode,
                    'hardwareEnabled': True,
                    'softwareEnabled': False
                }
                pallette = {'xpath': capture['xpath'] + '/filterPallette'}
                filter = {'xpath': capture['xpath'] + '/filter'}
                trigger = {'xpath': capture['xpath'] + '/trigger'}
                if len(capture_item.filters) > 0:
                    filter['captureFilterEnable'] = True
                    trigger['captureTriggerEnable'] = True
                    filter['captureFilterEnable'] = True
                    for cap_filter in capture_item.filters:
                        if cap_filter.parent.choice == 'ethernet':
                            self._config_ethernet_pallette(cap_filter,
                                                           pallette,
                                                           trigger,
                                                           filter)
                        elif cap_filter.parent.choice == 'custom':
                            self._config_custom_pallete(cap_filter,
                                                        pallette,
                                                        trigger,
                                                        filter)
                        else:
                            self._config_missing_pallete(cap_filter,
                                                        pallette,
                                                        trigger,
                                                        filter)
                imports.append(capture)
                imports.append(pallette)
                imports.append(filter)
                imports.append(trigger)
        self._import(imports)

    def reset_capture_request(self):
        self._capture_request = None
    
    def _config_missing_pallete(self, cap_filter, pallette, trigger, filter):
        pallete_map = getattr(self, '_{0}_OFFSET_MAP'.format(
                                cap_filter.parent.choice.upper()))
        for field_name in dir(cap_filter):
            if field_name not in pallete_map:
                raise Exception("Api not implimented for {0}".format(
                            field_name))
            if field_name in pallete_map:
                cap_field = getattr(cap_filter, field_name)
                if cap_field.value is not None:
                    new_pattern = GetPattern(filter.get(
                                        'captureFilterPattern'))
                    pallette[new_pattern.pattern] = cap_field.value
                    pallette[new_pattern.pattern_offset] = pallete_map[
                                field_name]
                    if cap_field.mask is not None:
                        pallette[new_pattern.pattern_mask] = cap_field.mask
                    if cap_field.negate is not None and \
                        cap_field.negate is True:
                        filter['captureFilterPattern'] = 'notPattern1'
                    else:
                        filter['captureFilterPattern'] = new_pattern.filter_pattern
                    trigger['triggerFilterPattern'] = filter[
                        'captureFilterPattern']
                
                    
    def _config_custom_pallete(self, cap_filter, pallette, trigger, filter):
        if cap_filter.value is not None:
            pallette['pattern1'] = cap_filter.value
            if cap_filter.mask is not None:
                pallette['patternMask1'] = cap_filter.mask
            if cap_filter.offset is not None:
                pallette['patternOffset1'] = cap_filter.offset
            if cap_filter.negate is not None and \
                    cap_filter.negate is True:
                filter['captureFilterPattern'] = 'notPattern1'
            else:
                filter['captureFilterPattern'] = 'pattern1'
            trigger['triggerFilterPattern'] = filter[
                'captureFilterPattern']

    def _config_ethernet_pallette(self, cap_filter, pallette, trigger, filter):
        if cap_filter.src.value is not None:
            pallette['SA1'] = cap_filter.src.value
            if cap_filter.src.mask is not None:
                pallette['SAMask1'] = cap_filter.src.mask
            if cap_filter.src.negate is not None and \
                    cap_filter.src.negate is True:
                filter['captureFilterSA'] = 'notAddr1'
            else:
                filter['captureFilterSA'] = 'addr1'
            trigger['triggerFilterSA'] = filter[
                'captureFilterSA']
        if cap_filter.dst.value is not None:
            pallette['DA1'] = cap_filter.dst.value
            if cap_filter.dst.mask is not None:
                pallette['DAMask1'] = cap_filter.dst.mask
            if cap_filter.dst.negate is not None and \
                    cap_filter.dst.negate is True:
                filter['captureFilterDA'] = 'notAddr1'
            else:
                filter['captureFilterDA'] = 'addr1'
            trigger['triggerFilterDA'] = filter[
                'captureFilterDA']

    def set_capture_state(self, request):
        """Starts capture on all ports that have capture enabled.
        """
        self._capture_request = request
        if request.state == 'start':
            self._start_capture()
        elif request.state == 'stop':
            self._stop_capture()
    
    def _start_capture(self):
        with Timer(self._api, 'Captures start'):
            if self._capture_request is None:
                return
            ixn_vports = self._api.select_vports()
            if len(ixn_vports) == 0:
                raise Exception("Please configure port before start capture")
            ixn_cap_ports = [name for name, vport in ixn_vports.items()
                            if vport['capture']['hardwareEnabled'] is True]
            port_names = self._capture_request.port_names
            if port_names is None or \
                    len(set(ixn_cap_ports) ^ set(port_names)) == 0:
                payload = {'arg1': []}
                for vport in ixn_vports.values():
                    payload['arg1'].append(vport['href'])
                url = '%s/vport/operations/clearCaptureInfos' % \
                      self._api._ixnetwork.href
                self._api._request('POST', url, payload)
                self._api._ixnetwork.StartCapture()
            else:
                url = '%s/vport/capture/operations/start' % \
                      self._api._ixnetwork.href
                for vport_name, vport in ixn_vports.items():
                    if vport_name in port_names:
                        if vport['capture']['hardwareEnabled'] is False:
                            raise Exception("Please enable capture in %s before start capture"
                                            %vport_name)
                        payload = {
                            'arg1': '{0}/capture'.format(vport['href'])
                        }
                        try:
                            self._api._request('POST', url, payload)
                        except Exception:
                            pass

    def _stop_capture(self):
        with Timer(self._api, 'Captures stop'):
            ixn_vports = self._api.select_vports()
            if len(ixn_vports) == 0:
                raise Exception("Please configure port before stop capture")
            if self._capture_request.port_names:
                payload = {'arg1': []}
                for vport_name, vport in ixn_vports.items():
                    if vport_name in self._capture_request.port_names:
                        payload['arg1'].append(vport['href'])
                url = '%s/vport/operations/clearCaptureInfos' % \
                        self._api._ixnetwork.href
                self._api._request('POST', url, payload)
            else:
                self._api._ixnetwork.StopCapture()
                

    def results(self, request):
        """Gets capture file and returns it as a byte stream
        """
        with Timer(self._api, 'Captures stop'):
            capture = self._api._vport.find(Name=self._api.special_char(
                            request.port_name)).Capture
            capture.Stop('allTraffic')

            #   Internally setting max time_out to 90sec with 3sec polling interval.
            #   Todo: Need to discuss and incorporate time_out field within model
            retry_count = 30
            port_ready = True
            for x in range(retry_count):
                port_ready = True
                time.sleep(3)
                capture = self._api._vport.find(Name=self._api.special_char(
                                request.port_name)).Capture
                if capture.HardwareEnabled and capture.DataCaptureState == 'notReady':
                    port_ready = False
                    continue
                if capture.SoftwareEnabled and capture.ControlCaptureState == 'notReady':
                    port_ready = False
                    continue
                break
            if not port_ready:
                self._api.warning("Capture was not stopped for this port %s" %
                             (request.port_name))

        payload = {'arg1': [self._api._vport.href]}
        url = '%s/vport/operations/getCaptureInfos' % self._api._ixnetwork.href
        response = self._api._request('POST', url, payload)
        file_name = response['result'][0]['arg6']
        file_id = response['result'][0]['arg1']

        url = '%s/vport/operations/saveCaptureInfo' % self._api._ixnetwork.href
        payload = {'arg1': self._api._vport.href, 'arg2': file_id}
        self._api._request('POST', url, payload)

        url = '%s/vport/operations/releaseCapturePorts' % self._api._ixnetwork.href
        payload = {'arg1': [self._api._vport.href]}
        self._api._request('POST', url, payload)

        path = '%s/capture' % self._api._ixnetwork.Globals.PersistencePath
        url = '%s/files?absolute=%s&filename=%s.cap' % (self._api._ixnetwork.href,
                                                        path, file_name)
        pcap_file_bytes = self._api._request('GET', url)
        return io.BytesIO(pcap_file_bytes)
        
class GetPattern(object):
    """ This is validating captureFilterPattern and return expected patterns
    """
    def __init__(self, cap_filter_pattern):
        self._new_count = 1
        self.cap_filter_pattern = cap_filter_pattern
        self._validate_pattern(cap_filter_pattern)
    
    def _validate_pattern(self, cap_filter_pattern):
        if cap_filter_pattern is not None:
            self._new_count = 2
    
    @property
    def pattern(self):
        return 'pattern{0}'.format(self._new_count)
    
    @property
    def pattern_mask(self):
        return 'patternMask{0}'.format(self._new_count)
    
    @property
    def pattern_offset(self):
        return 'patternOffset{0}'.format(self._new_count)
    
    @property
    def filter_pattern(self):
        if self._new_count == 1:
            return 'pattern{0}'.format(self._new_count)
        else:
            return '{0}AndPattern{1}'.format(self.cap_filter_pattern,
                                             self._new_count)
        
        